const { app, BrowserWindow, ipcMain, Tray, Menu } = require('electron')
const path = require('node:path')
const os = require('os');
const fs = require('fs');
const axios = require('axios');
const StreamZip = require('node-stream-zip');
const proc = require('child_process');
const AuthServer = require('./server');
const DEFAULT_PATH = path.join(os.homedir(), "eratz-server");
const API_URL = "https://play-eratz.com/api";
const tools = require('./utils');

const ZAAP_PORT = 26117;

var __playWindow, __mainWindow, __tray, __traycontextMenu;


console.log = console.info = console.warn = console.error = () => { };

function isGameInstalled(game_path, game_version_hash) {
  // [TODO] check version hash
  try {
    return fs.existsSync(path.join(game_path, 'loader.swf'))
  }
  catch (error) {
    return false;
  }
}

function openUrl(_, url) {
  const openUrl = process.platform === 'win32'
    ? `start ${url}`
    : process.platform === 'darwin'
      ? `open ${url}`
      : `xdg-open ${url}`;

  proc.exec(openUrl);
}

function getPlayProcessPath(gamePath) {
  switch (os.platform()) {
    case 'win32':
      return path.join(gamePath, 'Dofus.exe');
    case 'darwin':
      let retroPath = path.join(gamePath, 'Dofus.app', 'Contents', 'MacOS', 'Dofus Retro');
      proc.execSync(`chmod +x "${retroPath}"`);
      return retroPath;
    case 'linux':
      return path.join(gamePath, 'DofusLauncher');
    default:
      throw new Error(`unsupported platform.`);
  }
}

function setupGameDirectory(game_path) {
  let config_file;
  switch (os.platform()) {
    case 'win32':
    case 'linux':
      config_file = path.join(game_path, 'config.xml');
      break;
    case 'darwin':
      config_file = path.join(game_path, 'Dofus.app', 'Contents', 'Resources', 'app', 'retroclient', 'config.xml');
      break;
    default:
      throw new Error(`unsupported platform.`);
  }
  fs.cpSync(path.join(__dirname, 'assets', "config.xml"), config_file);
}

function fetchVersions() {
  return axios({ url: `${API_URL}/versions.json`, method: 'GET', headers: { Accept: 'application/json' } }).then((res) => res.data);
}

function authenticate(_, credentials) {
  return axios({ url: `${API_URL}/login`, method: 'POST', headers: { Accept: 'application/json' }, data: credentials })
    .then((res) => res.data).catch(err => {
      if (err.response?.status == 401) {
        return 0;
      }
      return -1;
    })
    .then((response) => {
      if (typeof response == 'number') {
        return response;
      }
      addAccount(response);
      return response;
    });
}

function downloadFile(fileUrl, outputLocationPath) {
  const writer = fs.createWriteStream(outputLocationPath);
  return axios({
    method: 'get',
    url: fileUrl,
    responseType: 'stream',
  }).then(response => {
    return new Promise((resolve, reject) => {
      const totalLength = response.headers.get('Content-Length');
      let receivedLength = 0;

      response.data.on('data', (chunk) => {
        receivedLength += chunk.length;
        const percent = ((receivedLength / totalLength) * 100).toFixed(2);
        __mainWindow.webContents.send('install-progress', { percent, has_next: true });
      });

      response.data.pipe(writer);

      let error = null;

      writer.on('error', err => {
        error = err;
        writer.close();
        reject(err);
      });
      writer.on('close', () => {
        if (!error) {
          resolve(true);
        }
      });
    });
  });
}

function installVersion(version) {
  try {
    if (!fs.existsSync(DEFAULT_PATH)) {
      fs.mkdirSync(DEFAULT_PATH);
    }

    let zipPath = path.join(DEFAULT_PATH, `temp_${version}.zip`);
    let game_path = path.join(DEFAULT_PATH, 'game');

    if (fs.existsSync(zipPath)) {
      fs.rmSync(zipPath);
    }

    return downloadFile(`${API_URL}/versions/${version}-${os.platform()}-${os.arch()}`, zipPath).then(async () => {
      try {

        if (fs.existsSync(game_path)) {
          fs.rmSync(game_path, { recursive: true });
        }

        __mainWindow.webContents.send('install-progress', { percent: 100, has_next: true, title: 'Copie de fichiers...' });

        const zip = new StreamZip.async({ file: zipPath });

        await zip.extract(null, game_path);

        setupGameDirectory(game_path)

        fs.writeFileSync(path.join(DEFAULT_PATH, 'stable.txt'), version);

        console.warn(`version "${version}" installed successfully.`);

        __mainWindow.webContents.send('install-progress', { percent: 101, has_next: false });

        return {}
      }
      catch (error) {
        if (fs.existsSync(zipPath)) {
          fs.rmSync(zipPath);
        }
        console.warn(`failed to extract version game "${version}" : ${error.message} (${error.stack})...`);
        return { error: error.message };
      }
    });
  }
  catch (error) {
    console.warn(`failed to install version "${version}" : ${error.message}...`);
    return Promise.resolve({ error: error.message });
  }
}

function getWorkspace() {
  try {
    let current_version, accounts, game_path = path.join(DEFAULT_PATH, 'game');
    let initialized = fs.existsSync(path.join(DEFAULT_PATH, 'stable.txt'));
    if (fs.existsSync(DEFAULT_PATH)) {
      try {
        if (initialized) {
          const stable = fs.readFileSync(path.join(DEFAULT_PATH, 'stable.txt'), 'utf-8').split('\n');
          current_version = stable[0];
        }
        else {
          current_version = null;
        }
      }
      catch (error) {
        console.warn(`error reading stable file...`);
        if (fs.existsSync(path.join(DEFAULT_PATH, 'stable.txt'))) {
          fs.rmSync(path.join(DEFAULT_PATH, 'stable.txt'));
        }
        current_version = null;
      }
      try {
        accounts = JSON.parse(Buffer.from(fs.readFileSync(path.join(DEFAULT_PATH, 'keys'), 'utf-8'), 'hex'));
      }
      catch (error) {
        console.warn(`error reading accounts file...`);
        if (fs.existsSync(path.join(DEFAULT_PATH, 'keys'))) {
          fs.rmSync(path.join(DEFAULT_PATH, 'keys'));
        }
        accounts = [];
      }
    }
    else {
      fs.mkdirSync(DEFAULT_PATH);
      current_version = null;
      accounts = [];
    }
    return Promise.all([
      fetchVersions(),
    ])
      .then((data) => {
        const [versions] = data;
        const { current: fetch_version, game_hash } = versions;
        return {
          initialized,
          fetch_version,
          current_version,
          game_path,
          game_installed: isGameInstalled(game_path, game_hash),
          accounts,
          playing: __playWindow !== undefined,
        }
      })
      .catch(err => {
        console.log(err)
        console.warn(`failed to fetch versions: ${err.message}...`);
        return { error: err.message }
      })
  }
  catch (error) {
    console.warn(`failed to create workspace: ${error.message}...`);
    return Promise.resolve({ error: error.message });
  }
}

function createWindow() {
  __mainWindow = new BrowserWindow({
    width: 750,
    height: 550,
    resizable: false,
    autoHideMenuBar: true,
    icon: path.join(__dirname, 'icon.png'),
    titleBarStyle: "darwin" === os.platform() ? void 0 : "hidden",
    // frame: false,
    // transparent: true,
    webPreferences: {
      preload: path.join(__dirname, 'preload.js'),
    }
  });

  __mainWindow.setSkipTaskbar(true);

  __mainWindow.once('close', function (e) {
    if (__playWindow) {
      __mainWindow.hide();
      __traycontextMenu.items[0].enabled = true;
      return;
    }
    app.exit();
    process.exit();
  });

  __mainWindow.on('show', function () {
  })

  __mainWindow.on('hide', function () {
    __traycontextMenu.items[0].enabled = true;
  })

  __mainWindow.loadFile('index.html');

  // app.dock.show();

  // installVersion('0.0.0')
  // __mainWindow.webContents.openDevTools()
}

function addAccount(newAccount) {
  let accounts = [];
  let accountsPath = path.join(DEFAULT_PATH, 'keys');
  if (fs.existsSync(accountsPath)) {
    try {
      accounts = JSON.parse(Buffer.from(fs.readFileSync(accountsPath, 'utf-8'), 'hex'));
    }
    catch (error) {
      accounts = [];
    }
  }
  fs.writeFileSync(accountsPath, Buffer.from(JSON.stringify([newAccount]), 'utf-8').toString('hex'));
}

const onMacAppSchemeHandlerCalled = (event, url) => {
  let parsed = new URL(url);
  let code = parsed.searchParams.get('key');
  if (code == null) {
    return;
  }
  console.info(`logging in account: ${code}...`);
  return addAccount(code);
}

const onAppSchemeHandlerCalled = (event, commandLine) => {
  const url = commandLine[commandLine.length - 1];
  onMacAppSchemeHandlerCalled(event, url);
}

if (os.platform() == 'darwin') {
  app.on('open-url', onMacAppSchemeHandlerCalled);
} else {
  app.on('second-instance', onAppSchemeHandlerCalled);
}

app.on('window-all-closed', () => {
  if (process.platform !== 'darwin') {
    app.quit();
  }
})

function playWindowClosed() {
  __playWindow.window.off('close', playWindowClosed);
  __playWindow = undefined;
  __mainWindow && __mainWindow.webContents.send('play-window-closed');
  __traycontextMenu.items[1].enabled = true;
}

function runPlayWindow() {
  if (__playWindow === undefined) {
    return getWorkspace().then((workspace) => {

      console.log(workspace);

      // re-assert game files
      setupGameDirectory(workspace.game_path)

      // run child process
      let execPath = getPlayProcessPath(workspace.game_path);
      __playWindow = (function () {

        const window = proc.spawn(execPath,
          {
            env: {
              ZAAP_HEROES: JSON.stringify(workspace.accounts),
              ...(
                {
                  ZAAP_CAN_AUTH: true,
                  ZAAP_GAME: "retro",
                  ZAAP_HASH: "",
                  ZAAP_INSTANCE_ID: "",
                  ZAAP_LOGS_PATH: path.join(workspace.game_path, 'logs'),
                  ZAAP_PORT: `${ZAAP_PORT}`,
                  ZAAP_RELEASE: "main"
                }
              )
            }
          }
        );

        console.log('game is running...');

        window.on('close', playWindowClosed);

        function close() {
          return window.kill();
        }

        return {
          close,
          window
        }

      })()
      __mainWindow.hide();
      __traycontextMenu.items[1].enabled = false;
      __traycontextMenu.items[0].enabled = true;
    })
  }
  __mainWindow.hide();
  return Promise.resolve();
}

app.whenReady().then(() => {
  __tray = new Tray(require('./logo'));

  __traycontextMenu = Menu.buildFromTemplate([
    {
      label: 'Show', type: 'normal', enabled: false, click: (menuItem, window, e) => {
        //createWindow();
        __mainWindow.show();
        __traycontextMenu.items[0].enabled = false;
      }
    },
    {
      label: 'Quit', type: 'normal', click: (menuItem, window, e) => {
        app.quit();
        process.exit(0);
      }
    },
  ])

  __tray.setToolTip('This is my application.');
  __tray.setContextMenu(__traycontextMenu)

  AuthServer(ZAAP_PORT, getWorkspace);

  const success = app.setAsDefaultProtocolClient('crifus');

  console.log('Protocol handler set:', success);

  ipcMain.handle('workspace', getWorkspace);
  ipcMain.handle('authenticate', authenticate);
  ipcMain.handle('play', runPlayWindow);
  ipcMain.handle('install', (_, version) => installVersion(version));
  ipcMain.handle('open-url', openUrl);

  createWindow();

  app.on('activate', function () {
    if (BrowserWindow.getAllWindows().length === 0) createWindow()
  })
})

app.on('window-all-closed', function () {
  if (process.platform !== 'darwin') app.quit()
})